@@ -1,6 +1,7 @@
 from celery import task, shared_task
 from .models import *
 from django.utils import timezone
+from django.db.models import Q
 
 import logging
 import pytz
@@ -46,88 +47,121 @@ def update_affiliations():
             logger.exception(e)
 
 @shared_task
-def update_characters():
+def update_tokens():
+    EveToken.objects.filter(evecharacter=None).delete()
+    for token in EveToken.objects.all():
+        token.refresh()
+
+    for token in EveToken.objects.all():
+        if token.valid and token.invalidated:
+            token.invalidated = None 
+            token.save() 
+        elif not token.valid and not token.invalidated:
+            token.invalidated = timezone.now()
+            token.save()
+        elif token.invalidated:
+            time_passed = timezone.now() - token.invalidated
+            if time_passed.days > 7:
+                token.delete()
+
+@shared_task
+def update_characters(jitter_max=1800):
     for eve_character in EveCharacter.objects.all():
-        if eve_character.token: 
-            update_character_assets.apply_async(args=[eve_character.external_id])
-            update_character_contacts.apply_async(args=[eve_character.external_id])
-            update_character_contracts.apply_async(args=[eve_character.external_id])
+        if eve_character.token and eve_character.token.valid and eve_character.corporation.track_characters: 
+            logger.info(f"Queueing batch update tasks for {eve_character.name}")
+            jitter = (eve_character.pk*10) % jitter_max
+            update_character_assets.apply_async(
+                args=[eve_character.external_id],
+                countdown=jitter)
+                
+            update_character_contacts.apply_async(
+                args=[eve_character.external_id],
+                countdown=jitter)
+                
+            update_character_contracts.apply_async(
+                args=[eve_character.external_id],
+                countdown=jitter)
+                
             update_character_journal.apply_async(
-                args=[eve_character.external_id])
+                args=[eve_character.external_id],
+                countdown=jitter)
+                
             update_character_jumpclones.apply_async(
-                args=[eve_character.external_id])
+                args=[eve_character.external_id],
+                countdown=jitter)
+                
             update_character_skills.apply_async(
-                args=[eve_character.external_id])
+                args=[eve_character.external_id],
+                countdown=jitter)
+                
             update_character_transactions.apply_async(
-                args=[eve_character.external_id])
+                args=[eve_character.external_id],
+                countdown=jitter)
+                
 
 @shared_task
 def update_corporations():
     for eve_corporation in EveCorporation.objects.all():
-        update_corporation_alliance.apply_async(
-            args=[eve_corporation.external_id])
-        update_corporation_ceo.apply_async(
-            args=[eve_corporation.external_id])
+        if eve_corporation.evecharacter_set.all().count() == 0:
+            eve_corporation.delete()
+            continue
+        if eve_corporation.track_corporation:
+            pull_corporation_roster.apply_async(args=[eve_corporation.external_id])
+            update_corporation_alliance.apply_async(
+                args=[eve_corporation.external_id])
+            update_corporation_ceo.apply_async(
+                args=[eve_corporation.external_id])
+        else:
+            logger.info(f"Skipping corporation update for {eve_corporation.name}: Not Tracked")
 
 @shared_task
 def update_alliances():
     for eve_alliance in EveAlliance.objects.all():
         update_alliance_executor.apply_async(args=[eve_alliance.external_id])
 
+@shared_task
+def update_structures():
+    for eve_corporation in EveCorporation.objects.filter(track_corporation=True):
+        try:
+            update_corporation_eveentitydata(
+                'corporations_corporation_id_structures', EveStructure, eve_corporation.external_id)
+        except Exception as e:
+            logger.error(e)
 """
 Character tasks 
 """
-def update_character_eveentitydata(op, *args, delete=True, **kwargs):
+def update_character_eveentitydata(op, data_model, character_id, delete=True):
     """
     Helper method for update_character_??? tasks.
     They basically all follow the same behavior. 
     """
-    if 'data_model' not in kwargs:
-        raise Exception("Must pass an EveEntityData model")
+    character = EveCharacter.objects.get(external_id=character_id)
 
-    data_model = kwargs.get('data_model')
+    response = EveClient.call(op, character_id=character.external_id)
 
-    if 'character_id' in kwargs:
-        characters = EveCharacter.objects.filter(
-            external_id=kwargs.get('character_id'))
-    else:
-        characters = []
-        tokens = EveToken.objects.filter(
-            scopes__in=EveClient.get_required_scopes(op))
-        for token in tokens:
-            characters.append(token.evecharacter)
+    if response.status != 200:
+        logger.error(f"Failed to batch update {data_model} for {character_id}: {response.header} {response.data}")
+        return
 
-    for character in characters:
-        try:
-            response = EveClient.call(op, character_id=character.external_id)
-        except Exception as e:
-            logger.exception(e)
-            continue
+    items = response.data
 
-        if response.status != 200:
-            logger.error(response)
-            continue
+    if 'X-Pages' in response.header and response.header['X-Pages'][0] > 1:
+        for num in range(2, response.header['X-Pages'][0]+1):
+            response = EveClient.call(
+                op, character_id=character.external_id, page=num)
+            if response.status != 200:
+                continue 
+            items += response.data 
 
-        items = response.data
+    if len(items) == 0:
+        return []
 
-        if 'X-Pages' in response.header and response.header['X-Pages'][0] > 1:
-            for num in range(2, response.header['X-Pages'][0]+1):
-                response = EveClient.call(
-                    op, character_id=character.external_id, page=num)
-                if response.status != 200:
-                    logger.error(response)
-                    continue 
-                items += response.data 
+    if delete:
+        data_model.objects.filter(entity=character).delete()
 
-        if len(items) == 0:
-            return []
-
-        if delete:
-            data_model.objects.filter(entity=character).delete()
-
-        data_model.create_from_esi_response(items, character.external_id)
-        
-        return items 
+    data_model.create_from_esi_response(items, character.external_id)
+    
+    return items 
 
 @shared_task
 def update_character_assets(character_id, *args, **kwargs):
@@ -186,6 +220,47 @@ def update_character_transactions(character_id, *args, **kwargs):
     update_character_eveentitydata(
         op, *args, delete=False, **kwargs, character_id=character_id, data_model=data_model)
 
+"""
+Corporatoin Tasks 
+"""
+
+
+def update_corporation_eveentitydata(op, data_model, corporation_id, delete=True):
+    """
+    Helper method for update_corporation_??? tasks.
+    They basically all follow the same behavior. 
+    """
+
+    corporation = EveCorporation.objects.get(external_id=corporation_id)
+
+    response = EveClient.call(op, corporation_id=corporation.external_id)
+
+    if response.status != 200:
+        logger.error(
+            f"Failed to batch update {data_model} for {character_id}: {response.header} {response.data}")
+        return
+
+
+    items = response.data
+
+    if 'X-Pages' in response.header and response.header['X-Pages'][0] > 1:
+        for num in range(2, response.header['X-Pages'][0]+1):
+            response = EveClient.call(
+                op, corporation_id=corporation.external_id, page=num)
+            if response.status != 200:
+                continue
+            items += response.data
+
+    if len(items) == 0:
+        return []
+
+    if delete:
+        data_model.objects.filter(entity=corporation).delete()
+
+    data_model.create_from_esi_response(items, corporation.external_id)
+
+    return items
+
 """
 Helper Tasks
 """
@@ -226,7 +301,7 @@ def update_corporation_ceo(corporation_id):
     if EveCharacter.objects.filter(external_id=ceo_id).exists():
         eve_corporation.ceo = EveCharacter.objects.get(external_id=ceo_id)
     else:
-        eve_corporation.ceo = EveCharacter.create_from_external_id(ceo_id)
+       logger.info(f"Skipping CEO update for {eve_corporation.external_id}: CEO not in database")
 
     eve_corporation.save()
 
@@ -234,11 +309,17 @@ def update_corporation_ceo(corporation_id):
 def pull_corporation_roster(corporation_id):
     token = EveToken.objects.filter(evecharacter__corporation__external_id=corporation_id).first()
     roster = EveClient.call('get_corporations_corporation_id_members', token=token, corporation_id=corporation_id)
-    for character in roster:
-        try:
+    members_to_update = EveCharacter.objects.filter(~Q(external_id__in=roster.data))
+    ids_that_exist = EveCharacter.objects.filter(external_id__in=roster.data).values_list('external_id', flat=True)
+
+    for character in roster.data:
+        if character not in ids_that_exist:
             EveCharacter.create_from_external_id(character)
-        except Exception as e:
-            logger.warning("Skipping %s in corporation roster" % character)
+        else:
+            logger.info(f"skipping {character} due to already existing")
+
+    for member in members_to_update:
+        member.update_character_corporation()
 
 @shared_task
 def update_alliance_executor(alliance_id):
